/* scheduler.cc
 * Copyright (C) 2001 The University of Newcastle upon Tyne
 *
 * Written by Paul Coates (Paul.Coates@ncl.ac.uk)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <gtk/gtk.h>
#include <arpa/inet.h>
#include <pwd.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>
#include <stdio.h>
#include <netdb.h>
#include "preferences.h"
#include "scheduler.h"
#include "topology.h"
#include "discover.h"
#include "process.h"
#include "task.h"
#include "main.h"
#include "session.h"
#include "data.h"
#include "misc.h"
#include "gui.h"
#include "monitor.h"
#include "tasklist.h"
#include "eventlog.h"

Scheduler::Scheduler(Main *m)
{
	main = m;

	discover1 = new Process();
	discover1->create(1);
	discover2 = new Process();
	discover2->create(1);
	discover3 = new Process();
	discover3->create(1);
	alive = new Process();
	alive->create(2);

	tasks.empty();

	ping_interval = MIN_PING_DEVICES_TIME;
	topology_interval = MIN_REGRAB_TOPOLOGY_TIME;
	autosave_interval = MIN_SAVE_SESSION_TIME;
	stats_interval = MIN_STATS_TIME;
	import_interval = MIN_IMPORT_TIME;
	ping_timer.now();
	topology_timer.now();
	autosave_timer.now();
	stats_timer.now();
	import_timer.now();
}

Scheduler::~Scheduler()
{
	Link *ln;
	Task *t;

	for (ln=tasks.get_head(); ln!=NULL; ln=ln->next)
	{ t = (Task*) ln->item; if (t!=NULL) delete t; }
	tasks.empty();

	alive->type = 0;
	alive->send("REQUEST");
	discover1->type = 0;
	discover1->send("REQUEST");
	discover2->type = 0;
	discover2->send("REQUEST");
	discover3->type = 0;
	discover3->send("REQUEST");
}

void Scheduler::addTask(char *str, int type, int lvl, int hop)
{
	Task *t;
	Link *ln;
	int match;

//printf("addTask(%s, %d, %d, %d)\n", str, type, lvl, hop);

	// check for duplicate tasks first
	match = 0;
	for (ln=tasks.get_head(); ln!=NULL; ln=ln->next)
	{
		t = (Task*) ln->item;
		if ((t->type==1) && (type==1) && (strcmp(str, t->cmnd)==0))
		{
			if (t->priority < lvl)
			{
				printf("A task with better priority than existing version\n");
				t->priority = lvl;
				main->gui->tasklist->update();
			}
			match = 1;
		}else
		if ((t->type==2) && (type==2))
		{
			// we already have a ping task
			match = 1;
		}
	}

	if (match==0)
	{
		t = new Task();
		strcpy(t->cmnd, str);
		t->priority = lvl;
		t->type = type;
		t->hop = hop;
		tasks.tag(t);
		human_message(t->cmnd, type);
		main->gui->tasklist->update();
		main->gui->monitor->show_task_label(tasks.size());
	}
}

void Scheduler::delTasks(int type)
{
	Link *ln;
	Task *t;

	for (ln=tasks.get_head(); ln!=NULL; ln=ln->next)
	{
		t = (Task*) ln->item;

		if (t->type == type)
		{
			delete t;
			ln->item = NULL;
		}
	}
	tasks.tidy();
	main->gui->monitor->show_task_label(tasks.size());
}

void Scheduler::returnedTask(Process *p)
{
	char buf[BUFSIZE];
	Link *ln2;
	Task *t;

	if (p->get(buf))
	{
//		printf("Process has completed (%d)\n", p->pid);
		// keep alive are not assigned to tasks
		if (p->type==3) p->type = 1;
		for (ln2=tasks.get_head(); ln2!=NULL; ln2=ln2->next)
		{
			t = (Task*) ln2->item;
			if (t->pid == p->pid)
			{
				if (p->type == 1)	// topology
				{
					main->topology->deviceAdd(p->param, t->hop);
					// write devices file to include new device
					main->session->save_device_file();
				}else
				if (p->type == 2) // alive
				{
					// do ping stuff
					pingcode();
				}
				delete t;
				ln2->item = NULL;
			}
		}
		tasks.tidy();
		main->gui->monitor->show_task_label(tasks.size());
		main->gui->tasklist->update();
		human_message("", 0);
		p->tid = 0;
		strcpy(p->param, "");
	}
}

void Scheduler::assignTask(Process *p)
{
	Link *ln2;
	Task *t;

	if (p->tid==0)
	{
		ln2=tasks.get_head();
		while(ln2!=NULL)
		{
			t = (Task*) ln2->item;
			if ((t->pid==0) && (t->type==p->type))
			{
				t->pid = p->pid;
				p->tid = 1;
				strcpy(p->param, t->cmnd);
				p->send("REQUEST");
				if ((t->type==1) && (t->priority>1))
				{
					human_message(t->cmnd, 3);
				}
				ln2=NULL;
				main->gui->tasklist->update();
			}else
			{
				ln2=ln2->next;
			}
		}
	}
}

void Scheduler::keepaliveTask(Process *p)
{
	if (p->tid!=0) return;
	p->type = 3;
	p->tid = 1;
	p->send("REQUEST");
}

// called every two seconds by main loop

void Scheduler::timeout()
{
	time_t now = time(NULL);

	// 1) check for processes ending and set appropriate flags
	returnedTask(alive);
	returnedTask(discover1);
	returnedTask(discover2);
	returnedTask(discover3);

//	alive->print();
//	discover1->print();
//	discover2->print();
//	discover3->print();

	// 2) check for new tasks

	if (ping_timer.ts + ping_interval <= now)
	{
		//	printf("Ping TIMER event\n");

		if ((main->session->devicebit==TRUE) ||
				(strcmp(main->session->session_name, "")==0))
			main->session->save_device_file();
		main->scheduler->addTask("", 2, 0, 0);
		ping_timer.ts = now;

		// also use ping timer to send keep alive messages
		keepaliveTask(discover1);
		keepaliveTask(discover2);
		keepaliveTask(discover3);
	}
	if (topology_timer.ts + topology_interval <= now)
	{
		// printf("Regrab topology TIMER event\n");
		// it's time to create a set of topology tasks
		main->scheduler->regrabAllDevices();
		topology_timer.ts = now;
	}
	if (autosave_timer.ts + autosave_interval <= now)
	{
		// printf("Auto save TIMER event\n");
		// if autosave set then save all files
		if (main->gui->preferences->autosave_flag==TRUE)
			main->session->save();
		autosave_timer.ts = now;
	}
	if (stats_timer.ts + stats_interval <= now)
	{
		// The server updates this often so we use the same time to reload
		// the summary file for marching ants, we also trigger refresh in
		// all graph windows as there is now more data  that can be viewed
		// printf("Update statistics TIMER event\n");
		// if autosave set then save all files
		stats_timer.ts = now;
	}
	if (import_timer.ts + import_interval <= now)
	{
		// printf("Import user data TIMER event\n");
		if (main->gui->preferences->import_auto_flag==TRUE)
			main->gui->preferences->import_import_action(FALSE);
		import_timer.ts = now;
	}

	// 3) assign tasks to processes
	assignTask(alive);
	assignTask(discover1);
	assignTask(discover2);
	assignTask(discover3);
}

/////////////////////////////////////////////////////////////////////////
// PING CODE
/////////////////////////////////////////////////////////////////////////

void Scheduler::pingcode()
{
	FILE *f;
	char buf[256];
	char buf2[256];
	Device *d;
	Link *ln;
	float ts;
	IPAddress tmpip;

	if (main->session->data==NULL) return;

// if status of any device changes here then log it and call notifier
// for now we just call the logger
// FIXME	main->gui->eventlog->add("enter message here");

	for (ln=main->session->data->devices.get_head(); ln!=NULL; ln=ln->next)
	{
		d = (Device*) ln->item;
		d->status = '0';
	}

	sprintf(buf, "%s/ping", main->session->datadir);
	f = fopen(buf, "r");
	if (!f) return;

	for (ln=main->session->data->devices.get_head(); ln!=NULL; ln=ln->next)
	{
		d = (Device*) ln->item;
		fseek(f, 0L, SEEK_SET);
		fscanf(f, "%[^\n]\n", buf); // the date stamp

		while(fscanf(f, "%[^\n]\n", buf)==1)
		{
			if (sscanf(buf, "%[^ ] (%f ms)", buf2, &ts)==2)
			{
				tmpip.setstring(buf2);
				if (tmpip.compare(d->pingname)==0)
					d->status = '1';
			}
		}
	}
	fclose(f);
	main->gui->update_views();
}

void Scheduler::human_message(char *str, int type)
{
	char buf[128];

	if (type==0)
	{
		strcpy(buf, "");
	}else
	if (type==1)
	{
		// Discovering device oakwell (128.240.0.114)
		sprintf(buf, "Discovering Topology From Device (%s)", str);
	}else
	if (type==2)
	{
		sprintf(buf, "Pinging Network Devices");
	}else
	if (type==3)
	{
		sprintf(buf, "Assigning Task (%s)", str);
	}else
		sprintf(buf, "Unknown task (%s)", str);

//	if (strlen(buf)>0) printf("%s\n", buf);
	main->gui->monitor->show_message(buf);
}

void Scheduler::regrabAllDevices()
{
	Link *ln;
	Device *d;

	for (ln=main->session->data->devices.get_head(); ln!=NULL; ln=ln->next)
	{
		d = (Device*) ln->item;
		if (d->grab_top=='1')
			regrabDevice(d);
	}
}

void Scheduler::regrabSelectedDevices()
{
	Link *ln;
	Device *d;

	for (ln=main->session->data->devices.get_head(); ln!=NULL; ln=ln->next)
	{
		d = (Device*) ln->item;
		if (d->selection==TRUE)
			regrabDevice(d);
	}
}

void Scheduler::regrabDevice(Device *d)
{
	gchar *buf, *csl;

	if (d==NULL) return;

	csl = get_community_list_string(main->client);
	buf = g_strjoin(" ", d->pingname.getstring(),
		(strcmp(d->community, "")==0)?csl:d->community, NULL);
	main->scheduler->addTask(buf, 1, 1, 0);
	g_free(buf);

	if (d->subtype==device_merged)
	{
		buf = g_strjoin(" ", d->pingname2.getstring(),
			(strcmp(d->community2, "")==0)?csl:d->community2, NULL);
		main->scheduler->addTask(buf, 1, 1, 0);
		g_free(buf);
	}
	g_free(csl);
}

